unit ubtree;

{$mode delphi}

interface

uses
  Classes, SysUtils;

type
  // Direcciones para navegación de un árblo binario
  TDirBTree = (DownIzq, DownDer, UpDer, UpIzq);


  TItemBTree = class
  end;


  { TNodoBtree }
  TNodoBtree = class
    izq, sup, der: TNodoBTree;
    item: TItemBTree;
    constructor Create(izq, sup, der: TNodoBTree; item: TItemBTree);
  end;


  { TBTree }
  TBTree = class
    raiz: TNodoBTree;
    flgAdmitirRepetidos: boolean;

    constructor Create(flgAdmitirRepetidos: boolean);

    // Compare impone el orden entre los individuos y debe ser definida
    // en una clase refinada de TBTree para poder usar esta clase
    //  -1 =  b < a ; 0 = b = a ; 1 = b > a
    function Compare(a, b: TItemBTree): integer; virtual; abstract;


    // resCod = 0, indica que fue insertado,
    // resCod = 2 es duplicado y no se inserto
    // en ambos casos retorna el Nodo del árbol.
    function insert(a: TItemBTree; out ResCod: integer): TNodoBTree; virtual;

    // busca el primer individuo tal que comparar( result, a ) = 0
    // y retorna el nodo o NIL si no lo encuentra.
    function locate_key(a: TItemBTree): TNodoBTree;

    // retorna el siguiente elemento del árbol en el sentido de dir y modifica dir
    // retorna NIL si no quedan elementos por visitar.
    // Cuando dir = UpDer o DownDer corresponde ejecutar la acción para la que se recorre
    // el arbol.
    procedure Next(var cursor: TNodoBTree; var dir: TDirBTree);

    // inicializa cursor = raiz y dir = DownDer
    procedure First(out cursor: TNodoBTree; out dir: TDirBTree);

  private

    // si resCod = -1 cursor = izq
    // si resCod = 0 cursor = self  (insertado)
    // si resCod = 1 cursor = der
    procedure insert_repetidos(var cursor: TNodoBTree; a: TItemBTree;
      var resCod: integer);

    // si resCod = -1 cursor = izq
    // si resCod = 0 cursor = self  (insertado)
    // si resCod = 1 cursor = der
    // si resCod = 2 cursor = self (duplicado - no insertado)
    function insert_unique(var cursor: TNodoBTree; a: TItemBTree;
      var resCod: integer): TNodoBTree;

  end;



implementation


{ TBTree }

constructor TBTree.Create(flgAdmitirRepetidos: boolean );
begin
  raiz := nil;
  self.flgAdmitirRepetidos := flgAdmitirRepetidos;
end;

function TBTree.insert(a: TItemBTree; out ResCod: integer): TNodoBTree;
var
  cursor: TNodoBtree;
begin
  if raiz = nil then
  begin
    raiz := TNodoBtree.Create(nil, nil, nil, a);
    resCod := 0;
    Result := raiz;
  end;

  cursor := raiz;
  if flgAdmitirRepetidos then
    repeat
      insert_repetidos(cursor, a, resCod);
    until abs(resCod) <> 1
  else
    repeat
      insert_unique(cursor, a, resCod);
    until abs(resCod) <> 1;
  Result := cursor;
end;


function TBTree.locate_key(a: TItemBTree): TNodoBTree;
var
  cursor: TNodoBTree;
  dir: TDirBTree;
  resComp: integer;
  buscando: boolean;
begin
  buscando := True;
  First(cursor, dir);
  while buscando and (cursor <> nil) do
  begin
    buscando := compare(a, cursor.item) <> 0;
    if buscando then
      Next(cursor, dir);
  end;
  Result := cursor;
end;

procedure TBTree.Next(var cursor: TNodoBTree; var dir: TDirBTree);
begin
  if cursor = nil then
  begin
    dir := UpIzq;
    exit;
  end;

  case dir of
    DownIzq:
    begin
      if cursor.izq <> nil then
        cursor := cursor.izq
      else if cursor.der <> nil then
      begin
        cursor := cursor.der;
        dir := DownDer;
      end
      else
      begin
        cursor := cursor.sup;
        dir := UpDer;
      end;
    end;
    UpDer:
    begin
      if cursor.der <> nil then
      begin
        cursor := cursor.der;
        dir := DownDer;
      end
      else
      begin
        cursor := cursor.sup;
        if (cursor.sup = nil) or (cursor.sup.der = cursor) then
          dir := UpIzq;
      end;
    end;
    DownDer:
    begin
      if cursor.izq <> nil then
      begin
        cursor := cursor.izq;
        dir := DownIzq;
      end
      else if cursor.der <> nil then
      begin
        cursor := cursor.der;
      end
      else
      begin
        cursor := cursor.sup;
        dir := UpIzq;
      end;
    end;
    else // UpIzq:
    begin
      cursor := cursor.sup;
      if (cursor.sup <> nil) then
        if cursor.sup.izq = cursor then
          dir := UpDer;
    end;
  end;
end;

procedure TBTree.First(out cursor: TNodoBTree; out dir: TDirBTree);
begin
  cursor := raiz;
  dir := DownDer;
end;

procedure TBTree.insert_repetidos(var cursor: TNodoBTree; a: TItemBTree;
  var resCod: integer);

var
  aNodo: TNodoBtree;
begin
  resCod := compare(cursor.item, a);
  case resCod of
    -1:
    begin
      if cursor.izq = nil then
      begin
        cursor.izq := TNodoBTree.Create(nil, cursor, nil, a);
        resCod := 0;
      end;
      cursor := cursor.izq;
    end;
    0:
    begin
      aNodo := TNodoBTree.Create(cursor.izq, cursor, nil, a);
      if cursor.der = nil then
      begin
        cursor.der := aNodo;
      end
      else
      begin
        aNodo.der := cursor.der;
        cursor.der.sup := aNodo;
        if cursor.izq <> nil then
          cursor.izq.sup := aNodo;
      end;
      cursor := aNodo;
    end;
    else // 1:
    begin
      if cursor.der = nil then
      begin
        cursor.der := TNodoBTree.Create(nil, cursor, nil, a);
        resCod := 0;
      end;
      cursor := cursor.der;
    end;
  end;
end;


function TBTree.insert_unique(var cursor: TNodoBTree; a: TItemBTree;
  var resCod: integer): TNodoBTree;
var
  aNodo: TNodoBtree;
begin

  resCod := compare(cursor.item, a);
  case resCod of
    -1:  // a < self
    begin
      if cursor.izq = nil then
      begin
        cursor.izq := TNodoBTree.Create(nil, cursor, nil, a);
        resCod := 0;
      end;
      cursor := cursor.izq;
    end;
    0: // a = self
    begin
      // cursor := cursor;
      resCod := 2;
    end;
    else // 1:
    begin  // a > self
      if cursor.der = nil then
        cursor.der := TNodoBTree.Create(nil, cursor, nil, a);
      resCod := 0;
      cursor := cursor.der;
    end;
  end;
end;

constructor TNodoBtree.Create(izq, sup, der: TNodoBTree; item: TItemBTree);
begin
  inherited Create;
  self.izq := izq;
  self.sup := sup;
  self.der := der;
  self.item := item;
end;



end.
